"""Option price computed with the Black'76 model"""
from jaxfin.price_engine.black.black_model import black_price, delta_black, gamma_black

future_option_price = black_price
future_option_delta = delta_black
future_option_gamma = gamma_black



"""
Black Scholes prices for Vanilla European options
"""
from typing import Callable, Optional, TypeVar, Union

import jax
import jax.numpy as jnp
from jax import grad, jit

from ..math.bs_common import compute_discounted_call_prices
from ..utils.arrays import cast_arrays

F = TypeVar("F", bound=Callable)


def bs_price(
    spots: jax.Array,
    strikes: jax.Array,
    expires: jax.Array,
    vols: jax.Array,
    discount_rates: jax.Array,
    are_calls: Optional[jax.Array] = None,
    dtype: Optional[jnp.dtype] = None,
) -> jax.Array:
    """
    Compute the option prices for european options using the Black-Scholes model.

    Args:
        spots (jax.Array): Array of current asset prices.
        strikes (jax.Array): Array of option strike prices.
        expires (jax.Array): Array of option expiration times.
        vols (jax.Array): Array of option volatility values.
        discount_rates (jax.Array): Array of risk-free interest rates. Defaults to None.
        dividend_rates (jax.Array, optional): Array of dividend rates. Defaults to None.
        are_calls (jax.Array): Array of booleans indicating whether options are calls (True) or puts (False).
        dtype (jnp.dtype, optional): Data type of the output. Defaults to None.

    Returns:
        jax.Array: Array of computed option prices.
    """
    [spots, strikes, expires, vols] = cast_arrays(
        [spots, strikes, expires, vols], dtype
    )

    discount_factors = jnp.exp(-discount_rates * expires)

    calls = compute_discounted_call_prices(
        spots, strikes, expires, vols, discount_rates
    )

    if are_calls is None:
        return calls

    puts = calls + (strikes * discount_factors) - spots
    return jnp.where(are_calls, calls, puts)


@jit
def delta_vanilla(
    spots: Union[jax.Array, float],
    strikes: Union[jax.Array, float],
    expires: Union[jax.Array, float],
    vols: Union[jax.Array, float],
    discount_rates: Union[jax.Array, float],
    are_calls: Optional[jax.Array] = None,
    dtype: Optional[jnp.dtype] = None,
) -> jax.Array:
    """
    Calculate the delta of a call/put option under the BS model (scalar version)

    :param spots: (Union[jax.Array, float]): Array of current asset prices.
    :param strikes: (Union[jax.Array, float]): Array of option strike prices.
    :param expires: (Union[jax.Array, float]): Array of option expiration times.
    :param vols: (Union[jax.Array, float]): Array of option volatility values.
    :param discount_rates: (Union[jax.Array, float]): Array of risk-free interest rates.
    :param are_calls: (jax.Array): Array of booleans indicating whether options
                                   are calls (True) or puts (False).
    :param dtype: (jnp.dtype): Data type of the output. Defaults to None.
    :return: (jax.Array): Array of delta of the given options.
    """
    return grad(bs_price, argnums=0)(
        spots, strikes, expires, vols, discount_rates, are_calls, dtype
    )


@jit
def gamma_vanilla(
    spots: Union[jax.Array, float],
    strikes: Union[jax.Array, float],
    expires: Union[jax.Array, float],
    vols: Union[jax.Array, float],
    discount_rates: Union[jax.Array, float],
    are_calls: Optional[jax.Array] = None,
    dtype: Optional[jnp.dtype] = None,
) -> jax.Array:
    """
    Calculate the gamma of a european option under the BS model (scalar version)

    :param spots: (Union[jax.Array, float]): Array of current asset prices.
    :param strikes: (Union[jax.Array, float]): Array of option strike prices.
    :param expires: (Union[jax.Array, float]): Array of option expiration times.
    :param vols: (Union[jax.Array, float]): Array of option volatility values.
    :param discount_rates: (Union[jax.Array, float]): Array of risk-free interest rates
    :param are_calls: (jax.Array) Array of booleans indicating whether options
                                  are calls (True) or puts (False).
    :param dtype: (jnp.dtype): Data type of the output. Defaults to None.
    :return: (jax.Array): Array of gamma of the given options.
    """
    return grad(grad(bs_price, argnums=0), argnums=0)(
        spots, strikes, expires, vols, discount_rates, are_calls, dtype
    )


@jit
def theta_vanilla(
    spots: Union[jax.Array, float],
    strikes: Union[jax.Array, float],
    expires: Union[jax.Array, float],
    vols: Union[jax.Array, float],
    discount_rates: Union[jax.Array, float],
    are_calls: Optional[jax.Array] = None,
    dtype: Optional[jnp.dtype] = None,
) -> jax.Array:
    """
    Calculate the theta of a european option under the BS model (scalar version)

    :param spots: (Union[jax.Array, float]): Array of current asset prices.
    :param strikes: (Union[jax.Array, float]): Array of option strike prices.
    :param expires: (Union[jax.Array, float]): Array of option expiration times.
    :param vols: (Union[jax.Array, float]): Array of option volatility values.
    :param discount_rates: (Union[jax.Array, float]): Array of risk-free interest rates
    :param are_calls: (jax.Array) Array of booleans indicating whether options
                                  are calls (True) or puts (False).
    :param dtype: (jnp.dtype): Data type of the output. Defaults to None.
    :return: (jax.Array): Array of theta of the given options.
    """
    return grad(bs_price, argnums=2)(
        spots, strikes, expires, vols, discount_rates, are_calls, dtype
    )


@jit
def vega_vanilla(
    spots: Union[jax.Array, float],
    strikes: Union[jax.Array, float],
    expires: Union[jax.Array, float],
    vols: Union[jax.Array, float],
    discount_rates: Union[jax.Array, float],
    are_calls: Optional[jax.Array] = None,
    dtype: Optional[jnp.dtype] = None,
) -> jax.Array:
    """
    Calculate the vega of a european option under the BS model (scalar version)

    :param spots: (Union[jax.Array, float]): Array of current asset prices.
    :param strikes: (Union[jax.Array, float]): Array of option strike prices.
    :param expires: (Union[jax.Array, float]): Array of option expiration times.
    :param vols: (Union[jax.Array, float]): Array of option volatility values.
    :param discount_rates: (Union[jax.Array, float]): Array of risk-free interest rates
    :param are_calls: (jax.Array) Array of booleans indicating whether options
                                  are calls (True) or puts (False).
    :param dtype: (jnp.dtype): Data type of the output. Defaults to None.
    :return: (jax.Array): Array of vega of the given options.
    """
    return grad(bs_price, argnums=3)(
        spots, strikes, expires, vols, discount_rates, are_calls, dtype
    )


@jit
def rho_vanilla(
    spots: Union[jax.Array, float],
    strikes: Union[jax.Array, float],
    expires: Union[jax.Array, float],
    vols: Union[jax.Array, float],
    discount_rates: Union[jax.Array, float],
    are_calls: Optional[jax.Array] = None,
    dtype: Optional[jnp.dtype] = None,
) -> jax.Array:
    """
    Calculate the rho of a european option under the BS model (scalar version)

    :param spots: (Union[jax.Array, float]): Array of current asset prices.
    :param strikes: (Union[jax.Array, float]): Array of option strike prices.
    :param expires: (Union[jax.Array, float]): Array of option expiration times.
    :param vols: (Union[jax.Array, float]): Array of option volatility values.
    :param discount_rates: (Union[jax.Array, float]): Array of risk-free interest rates
    :param are_calls: (jax.Array) Array of booleans indicating whether options
                                  are calls (True) or puts (False).
    :param dtype: (jnp.dtype): Data type of the output. Defaults to None.
    :return: (jax.Array): Array of rho of the given options.
    """
    return grad(bs_price, argnums=4)(
        spots, strikes, expires, vols, discount_rates, are_calls, dtype
    )



"""
Vectorization realted function used by the price engine module
"""
from typing import Callable, Optional, Tuple, TypeVar, Union

import jax
import jax.numpy as jnp
from jax import jit, vmap

F = TypeVar("F", bound=Callable)


def get_vfunction(
    fun: F,
    spots: Union[jax.Array, float],
    strikes: Union[jax.Array, float],
    expires: Union[jax.Array, float],
    vols: Union[jax.Array, float],
    discount_rates: Union[jax.Array, float],
    dividends_rates: Optional[Union[jax.Array, float]] = None,
    are_calls: Optional[Union[jax.Array, bool]] = None,
    dtype: Optional[jnp.dtype] = None,
):
    """
    Returns a vectorized function for the Black-Scholes related functions.

    Args:
        fun (Callable): Function to vectorize.
        spots (Union[jax.Array, float]): Current asset price or array of prices.
        strikes (Union[jax.Array, float]): Option strike price or array of prices.
        expires (Union[jax.Array, float]): Option expiration time or array of times.
        vols (Union[jax.Array, float]): Option volatility value or array of values.
        discount_rates (Union[jax.Array, float]): Risk-free interest rate or array of rates.
        dividend_rates (Union[jax.Array, float], optional): Dividend rate or array of rates. Defaults to None.
        are_calls (Union[jax.Array, bool], optional): Boolean indicating whether option is a
                                                      call or put, or array of booleans. Defaults to None.
        dtype (jnp.dtype, optional): Data type of the output. Defaults to None.

    Returns:
        Vectorized function.
    """
    return jit(
        vmap(
            fun,
            in_axes=_get_vmap_mask(
                spots,
                strikes,
                expires,
                vols,
                discount_rates,
                dividends_rates,
                are_calls,
                dtype,
            ),
        )
    )


def _get_vmap_mask(
    spots: Union[jax.Array, float],
    strikes: Union[jax.Array, float],
    expires: Union[jax.Array, float],
    vols: Union[jax.Array, float],
    discount_rates: Union[jax.Array, float],
    dividend_rates: Optional[Union[jax.Array, float]] = None,
    are_calls: Optional[Union[jax.Array, bool]] = None,
    dtype: Union[jnp.dtype, None] = None,
) -> Tuple[Union[None, int], ...]:
    mask = tuple(
        map(
            lambda x: None if jnp.isscalar(x) else 0,
            [spots, strikes, expires, vols, discount_rates],
        )
    )

    if dividend_rates is not None:
        mask += (0,)

    if are_calls is not None:
        mask += (0,)

        if dtype is not None:
            mask += (None,)

    return mask



"""
Heston model class implementation
"""
from typing import Optional, Union

import jax
import jax.numpy as jnp
from jax import jit, random


class UnivHestonModel:
    """
    Univariate Heston Model

    Represent a 1-dimensional Heston model

    # Example usage:
    s0 = jnp.array([100], dtype=jnp.float32)
    v0 = jnp.array([0.04], dtype=jnp.float32)
    kappa = jnp.array([2.0], dtype=jnp.float32)
    theta = jnp.array([0.04], dtype=jnp.float32)
    sigma = jnp.array([0.3], dtype=jnp.float32)
    rho = jnp.array([-0.7], dtype=jnp.float32)

    heston_model = UnivHestonModel(s0, v0, kappa, theta, sigma, rho, dtype=jnp.float32)
    paths = heston_model.simulate_paths(maturity=1.0, n=100, n_sim=100)
    """

    def __init__(
        self,
        s0: Union[jax.Array, float],
        v0: Union[jax.Array, float],
        mean: Union[jax.Array, float],
        kappa: Union[jax.Array, float],
        theta: Union[jax.Array, float],
        sigma: Union[jax.Array, float],
        rho: Union[jax.Array, float],
        dtype: Optional[jax.numpy.dtype] = jnp.float32,
    ):
        """Univariate Heston model constructor

        Args:
            s0 (Union[jax.Array, float]): The initial value of the asset
            v0 (Union[jax.Array, float]): The initial of the variance of the asset
            mean (Union[jax.Array, float]): The mean of the asset
            kappa (Union[jax.Array, float]): The speed of the mean-reversion of the variance
            theta (Union[jax.Array, float]): The long-term mean of the variance
            sigma (Union[jax.Array, float]): The volatility of the variance
            rho (Union[jax.Array, float]): The correlation between the asset and the variance
            dtype (Optional[jax.numpy.dtype], optional): The type(jnp.float32, ...). Defaults to jnp.float32.
        """
        if dtype is None:
            raise ValueError("dtype must not be None")

        self._dtype = dtype
        self._s0 = jnp.asarray(s0, dtype=dtype)
        self._v0 = jnp.asarray(v0, dtype=dtype)
        self._mean = jnp.asarray(mean, dtype=dtype)
        self._kappa = jnp.asarray(kappa, dtype=dtype)
        self._theta = jnp.asarray(theta, dtype=dtype)
        self._sigma = jnp.asarray(sigma, dtype=dtype)
        self._rho = jnp.asarray(rho, dtype=dtype)

    @property
    def mean(self) -> jax.Array:
        """
        :return: Returns the mean of the Heston model
        """
        return self._mean

    @property
    def kappa(self) -> jax.Array:
        """
        :return: Returns the speed of the mean-reversion of the variance
        """
        return self._kappa

    @property
    def theta(self) -> jax.Array:
        """
        :return: Returns the long-term mean of the variance
        """
        return self._theta

    @property
    def sigma(self) -> jax.Array:
        """
        :return: Returns the volatility of the variance
        """
        return self._sigma

    @property
    def rho(self) -> jax.Array:
        """
        :return: Returns the correlation between the asset and the variance
        """
        return self._rho

    @property
    def s0(self) -> jax.Array:
        """
        :return: Returns the initial value of the asset
        """
        return self._s0

    @property
    def v0(self) -> jax.Array:
        """
        :return: Returns the initial value of the variance of the asset
        """
        return self._v0

    @property
    def dtype(self) -> jax.numpy.dtype:
        """
        :return: Returns the underlying dtype of the Heston model
        """
        return self._dtype

    def sample_paths(self, seed: int, maturity: float, n: int, n_sim: int):
        """
        Simulate a sample of paths from the Heston model

        Args:
            seed (int): the seed for the random number generator
            maturity (float): The time in years
            n (int): Number of steps (discretization points)
            n_sim (int): The number of simulations
        """
        key = random.PRNGKey(seed)

        dt = maturity / n

        W = random.normal(key, shape=(n_sim, n)).T
        Z = (
            self._rho
            * W
            * jnp.sqrt(1 - self._rho**2)
            * random.normal(key, shape=(n_sim, n)).T
        )

        return _sample_paths(
            self.s0, self.v0, self.mean, self.kappa, self.theta, self.sigma, dt, W, Z, n
        )


def _variance_process_step(
    v: jax.Array,
    kappa: jax.Array,
    theta: jax.Array,
    sigma: jax.Array,
    dt: jax.Array,
    dZ: jax.Array,
) -> jax.Array:
    return (
        v
        + kappa * (theta - jnp.maximum(v, 0.0)) * dt
        + sigma * jnp.sqrt(jnp.maximum(v, 0.0)) * jnp.sqrt(dt) * dZ
    )


@jit
def _sample_paths(
    s0: jax.Array,
    v0: jax.Array,
    mean: jax.Array,
    kappa: jax.Array,
    theta: jax.Array,
    sigma: jax.Array,
    dt: jax.Array,
    W: jax.Array,
    Z: jax.Array,
    N: int,
) -> jax.Array:
    def init_fn(W, Z):
        W_ = W
        Z_ = Z

        S = jnp.full_like(W_, s0)
        v = jnp.full_like(W_, v0)

        return 0, S, v, W_, Z_

    def cond_fn(val):
        i, *_ = val
        return i < N

    def body_val(val):
        i, S, v, W, Z = val
        dW = W[i]
        dZ = Z[i]

        S = S.at[i + 1].set(
            S[i]
            * jnp.exp((mean - 0.5 * v[i]) * dt + jnp.sqrt(v[i]) * jnp.sqrt(dt) * dW)
        )

        v = v.at[i + 1].set(_variance_process_step(v[i], kappa, theta, sigma, dt, dZ))

        return i + 1, S, v, W, Z

    _, S, _, _, _ = jax.lax.while_loop(cond_fn, body_val, init_fn(W, Z))

    return S



"""
This module contains the models for the JAXFIN package.

The submodules of this module are:
    
        * gbm: Geometric Brownian Motion model
"""
from jaxfin.models import gbm

__all__ = ["gbm"]
